/*******************************************************************************

NAME            cpu_map.c
VERSION         %version:   1 %
UPDATE DATE     %date_modified: Sat Apr 30 12:09:47 2016 %
PROGRAMMER      %created_by: sneheetk %

    Copyright 2016 NetApp, Inc.

DESCRIPTION: Grub2 module to detect number of logical cores present and set
isolcpus kernel configuration parameter accordingly. The current implementation
makes core 0 and its hyperthreaded core available to the scheduler and reserves
remaining cores for RAID application.

LICENSE:
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*******************************************************************************/

/***  INCLUDES  ***/
#include <grub/dl.h>
#include <grub/env.h>
#include <grub/command.h>

/***  CONSTANT DEFINITIONS  ***/
GRUB_MOD_LICENSE ("GPLv3+");
GRUB_MOD_DUAL_LICENSE ("All Rights Reserved");

#define CPUID_LEAF (0xB)
#define NUM_LOGICAL_CORES_MASK (0xFFFF)
#define LEVEL_NUM_MASK (0xFF)

/*** MACRO DEFINITIONS ***/
#define DBG_PRT(arg...)  do { if (enableDbgPrints) grub_printf(arg); } while (0)

/***  LOCALS  ***/
static int enableDbgPrints = 0;

/*** PROCEDURES ***/

/**
 * Execute cpuid instruction. Input values are in eax and ecx registers. Output values are in eax,
 * ebx and ecx registers.
 */
static inline void
nativeCpuId
    (
    unsigned int *eax,
    unsigned int *ebx,
    unsigned int *ecx,
    unsigned int *edx
    )
{
    __asm__ volatile("cpuid"
        : "=a" (*eax), "=b" (*ebx), "=c" (*ecx), "=d" (*edx)
        : "0" (*eax), "2" (*ecx)
        : "memory");
}

/**
 * Execute cpuid instruction to get total number of logical cores.
 * NOTE - This function returns total number of logical cores even when hyperthreading is
 * disabled in the BIOS.
 */
static unsigned int
getNumLogicalCores
    (
    void
    )
{

    unsigned int eax = CPUID_LEAF;
    unsigned int ebx = 0;
    unsigned int ecx = 0;
    unsigned int edx = 0;
    unsigned int numCores = 0;

    /* Set input registers eax to 0xB and ecx to 0. Bits [7:0] of ecx contain level number.
     * CPUID instruction returns number of logical cores in bits [15:0] of ebx register at
     * the given level. Increment level number in ecx and loop until cpuid returns 0 in ebx.
     */
    while (1)
    {
        nativeCpuId(&eax, &ebx, &ecx, &edx);
        if (ebx == 0)
            break;
        numCores = ebx & NUM_LOGICAL_CORES_MASK;
        ecx++;
        ecx &= LEVEL_NUM_MASK;
        eax = CPUID_LEAF;
    }

    return numCores;
}

/**
 * The guts of the 'cpu_map' command.
 * Determine the platform and set isolcpus.
 */
static grub_err_t
grub_cmd_cpu_map
    (
    grub_command_t cmd __attribute__ ((unused)),
    int argc,
    char *argv[]
    )
{
    char *varName;
    unsigned int numCores;
    char buf[50];

    if (argc == 0 || argc > 1)
    {
        return grub_error(GRUB_ERR_BAD_ARGUMENT, N_("one argument expected"));
    }

    /* If "DEBUG" env var exists, enable prints */
    enableDbgPrints = grub_env_get("DEBUG") ? 1 : 0;

    varName = argv[0];
    DBG_PRT("variable: %s\n", varName);

    numCores = getNumLogicalCores();

    /* Note: cpuid instruction will not return correct number of logical cores when
     * hyperthreading is disabled. This code will still isolate the correct list of cores.
     * For e.g. 4-core CPU with HT off: numCores = 8. isolcpus=1-3,5-7. As only cores 0-3 are
     * online, isolcpus will correctly isolate cores 1-3, and ignore 5-7.
     * This code assumes that all our processors support hyperthreading. It will need to be
     * changed if we come across a processor that doesn't support hyperthreading.
     */
    if (numCores < 2)
        grub_snprintf(buf, sizeof(buf), " ");
    else if (numCores == 2)
        grub_snprintf(buf, sizeof(buf), "isolcpus=1");
    else if (numCores == 4)
        grub_snprintf(buf, sizeof(buf), "isolcpus=1,3");
    else
        grub_snprintf(buf, sizeof(buf), "isolcpus=%u-%u,%u-%u",
            1, numCores / 2 - 1, numCores / 2 + 1, numCores - 1);

    grub_env_set(varName, buf);

    return GRUB_ERR_NONE;
}

/* Register/unregister our command with module init/exit */
static grub_command_t cmd;

GRUB_MOD_INIT(cpu_map)
{
    cmd = grub_register_command("cpu_map", grub_cmd_cpu_map, N_("ENV_VAR"),
              N_("Sets isolcpus kernel config parameter to isolate CPU cores from scheduler."));
}

GRUB_MOD_FINI(cpu_map)
{
    if (cmd)
    {
        grub_unregister_command(cmd);
        cmd = 0;
    }
}
